@@ -31,7 +31,7 @@
 
 #pragma once
 
-#include "ts/ink_sock.h"
+#include "tscore/ink_sock.h"
 #include "I_NetVConnection.h"
 #include "P_UnixNetState.h"
 #include "P_Connection.h"
@@ -67,6 +67,7 @@ NetVCOptions::reset()
   etype = ET_NET;
 
   sni_servername    = nullptr;
+  ssl_servername    = nullptr;
   clientCertificate = nullptr;
 }
 
@@ -90,7 +91,7 @@ struct OOB_callback : public Continuation {
   int retry_OOB_send(int, Event *);
 
   OOB_callback(Ptr<ProxyMutex> &m, NetVConnection *vc, Continuation *cont, char *buf, int len)
-    : Continuation(m), data(buf), length(len), trigger(0)
+    : Continuation(m), data(buf), length(len), trigger(nullptr)
   {
     server_vc   = (UnixNetVConnection *)vc;
     server_cont = cont;
@@ -154,7 +155,6 @@ public:
   void cancel_active_timeout() override;
   void cancel_inactivity_timeout() override;
   void set_action(Continuation *c) override;
-  const Action *get_action() const;
   void add_to_keep_alive_queue() override;
   void remove_from_keep_alive_queue() override;
   bool add_to_active_queue() override;
@@ -166,7 +166,7 @@ public:
 
   SOCKET get_socket() override;
 
-  virtual ~UnixNetVConnection();
+  ~UnixNetVConnection() override;
 
   /////////////////////////////////////////////////////////////////
   // instances of UnixNetVConnection should be allocated         //
@@ -175,39 +175,13 @@ public:
   /////////////////////////////////////////////////////////////////
   UnixNetVConnection();
 
-  int
-  populate_protocol(ts::StringView *results, int n) const override
-  {
-    int retval = 0;
-    if (n > retval) {
-      if (!(results[retval] = options.get_proto_string()).isEmpty())
-        ++retval;
-      if (n > retval) {
-        if (!(results[retval] = options.get_family_string()).isEmpty())
-          ++retval;
-      }
-    }
-    return retval;
-  }
+  int populate_protocol(std::string_view *results, int n) const override;
+  const char *protocol_contains(std::string_view tag) const override;
 
-  const char *
-  protocol_contains(ts::StringView tag) const override
-  {
-    ts::StringView retval = options.get_proto_string();
-    if (strncmp(tag.ptr(), retval.ptr(), tag.size()) != 0) {
-      retval = options.get_family_string();
-      if (strncmp(tag.ptr(), retval.ptr(), tag.size()) != 0) {
-        retval.clear();
-      }
-    }
-    return retval.ptr();
-  }
+  // noncopyable
+  UnixNetVConnection(const NetVConnection &) = delete;
+  UnixNetVConnection &operator=(const NetVConnection &) = delete;
 
-private:
-  UnixNetVConnection(const NetVConnection &);
-  UnixNetVConnection &operator=(const NetVConnection &);
-
-public:
   /////////////////////////
   // UNIX implementation //
   /////////////////////////
@@ -232,12 +206,6 @@ public:
     return (true);
   }
 
-  virtual bool
-  trackFirstHandshake()
-  {
-    return false;
-  }
-
   virtual void net_read_io(NetHandler *nh, EThread *lthread);
   virtual int64_t load_buffer_and_write(int64_t towrite, MIOBufferAccessor &buf, int64_t &total_written, int &needs);
   void readDisable(NetHandler *nh);
@@ -255,7 +223,7 @@ public:
   UnixNetVConnection *migrateToCurrentThread(Continuation *c, EThread *t);
 
   Action action_;
-  volatile int closed;
+  int closed;
   NetState read;
   NetState write;
 
@@ -269,13 +237,8 @@ public:
 
   ink_hrtime inactivity_timeout_in;
   ink_hrtime active_timeout_in;
-#ifdef INACTIVITY_TIMEOUT
-  Event *inactivity_timeout;
-  Event *activity_timeout;
-#else
   ink_hrtime next_inactivity_timeout_at;
   ink_hrtime next_activity_timeout_at;
-#endif
 
   EventIO ep;
   NetHandler *nh;
@@ -303,7 +266,6 @@ public:
   const sockaddr *origin_trace_addr;
   int origin_trace_port;
 
-  int startEvent(int event, Event *e);
   int acceptEvent(int event, Event *e);
   int mainEvent(int event, Event *e);
   virtual int connectUp(EThread *t, int fd);
@@ -313,13 +275,16 @@ public:
    * This is logic is invoked when the NetVC object is created in a new thread context
    */
   virtual int populate(Connection &con, Continuation *c, void *arg);
+  virtual void clear();
   virtual void free(EThread *t);
 
   ink_hrtime get_inactivity_timeout() override;
   ink_hrtime get_active_timeout() override;
 
   void set_local_addr() override;
+  void set_mptcp_state() override;
   void set_remote_addr() override;
+  void set_remote_addr(const sockaddr *) override;
   int set_tcp_init_cwnd(int init_cwnd) override;
   int set_tcp_congestion_control(int side) override;
   void apply_options() override;
@@ -353,6 +318,14 @@ TS_INLINE void
 UnixNetVConnection::set_remote_addr()
 {
   ats_ip_copy(&remote_addr, &con.addr);
+  this->control_flags.set_flag(ContFlags::DEBUG_OVERRIDE, diags->test_override_ip(remote_addr));
+}
+
+TS_INLINE void
+UnixNetVConnection::set_remote_addr(const sockaddr *new_sa)
+{
+  ats_ip_copy(&remote_addr, new_sa);
+  this->control_flags.set_flag(ContFlags::DEBUG_OVERRIDE, diags->test_override_ip(remote_addr));
 }
 
 TS_INLINE void
@@ -365,6 +338,21 @@ UnixNetVConnection::set_local_addr()
   ATS_UNUSED_RETURN(safe_getsockname(con.fd, &local_addr.sa, &local_sa_size));
 }
 
+// Update the internal VC state variable for MPTCP
+TS_INLINE void
+UnixNetVConnection::set_mptcp_state()
+{
+  int mptcp_enabled      = -1;
+  int mptcp_enabled_size = sizeof(mptcp_enabled);
+
+  if (0 == safe_getsockopt(con.fd, IPPROTO_TCP, MPTCP_ENABLED, (char *)&mptcp_enabled, &mptcp_enabled_size)) {
+    Debug("socket_mptcp", "MPTCP socket state: %d", mptcp_enabled);
+    mptcp_state = mptcp_enabled > 0 ? true : false;
+  } else {
+    Debug("socket_mptcp", "MPTCP failed getsockopt(): %s", strerror(errno));
+  }
+}
+
 TS_INLINE ink_hrtime
 UnixNetVConnection::get_active_timeout()
 {
@@ -381,30 +369,8 @@ TS_INLINE void
 UnixNetVConnection::set_active_timeout(ink_hrtime timeout_in)
 {
   Debug("socket", "Set active timeout=%" PRId64 ", NetVC=%p", timeout_in, this);
-  active_timeout_in = timeout_in;
-#ifdef INACTIVITY_TIMEOUT
-  if (active_timeout)
-    active_timeout->cancel_action(this);
-  if (active_timeout_in) {
-    if (read.enabled) {
-      ink_assert(read.vio.mutex->thread_holding == this_ethread() && thread);
-      if (read.vio.mutex->thread_holding == thread)
-        active_timeout = thread->schedule_in_local(this, active_timeout_in);
-      else
-        active_timeout = thread->schedule_in(this, active_timeout_in);
-    } else if (write.enabled) {
-      ink_assert(write.vio.mutex->thread_holding == this_ethread() && thread);
-      if (write.vio.mutex->thread_holding == thread)
-        active_timeout = thread->schedule_in_local(this, active_timeout_in);
-      else
-        active_timeout = thread->schedule_in(this, active_timeout_in);
-    } else
-      active_timeout = 0;
-  } else
-    active_timeout = 0;
-#else
-  next_activity_timeout_at = Thread::get_hrtime() + timeout_in;
-#endif
+  active_timeout_in        = timeout_in;
+  next_activity_timeout_at = (active_timeout_in > 0) ? Thread::get_hrtime() + timeout_in : 0;
 }
 
 TS_INLINE void
@@ -412,31 +378,15 @@ UnixNetVConnection::cancel_inactivity_timeout()
 {
   Debug("socket", "Cancel inactive timeout for NetVC=%p", this);
   inactivity_timeout_in = 0;
-#ifdef INACTIVITY_TIMEOUT
-  if (inactivity_timeout) {
-    Debug("socket", "Cancel inactive timeout for NetVC=%p", this);
-    inactivity_timeout->cancel_action(this);
-    inactivity_timeout = nullptr;
-  }
-#else
   set_inactivity_timeout(0);
-#endif
 }
 
 TS_INLINE void
 UnixNetVConnection::cancel_active_timeout()
 {
   Debug("socket", "Cancel active timeout for NetVC=%p", this);
-  active_timeout_in = 0;
-#ifdef INACTIVITY_TIMEOUT
-  if (active_timeout) {
-    Debug("socket", "Cancel active timeout for NetVC=%p", this);
-    active_timeout->cancel_action(this);
-    active_timeout = nullptr;
-  }
-#else
+  active_timeout_in        = 0;
   next_activity_timeout_at = 0;
-#endif
 }
 
 TS_INLINE int
@@ -454,9 +404,7 @@ UnixNetVConnection::set_tcp_init_cwnd(int init_cwnd)
 #endif
 }
 
-TS_INLINE UnixNetVConnection::~UnixNetVConnection()
-{
-}
+TS_INLINE UnixNetVConnection::~UnixNetVConnection() {}
 
 TS_INLINE SOCKET
 UnixNetVConnection::get_socket()
@@ -470,14 +418,7 @@ UnixNetVConnection::set_action(Continuation *c)
   action_ = c;
 }
 
-TS_INLINE const Action *
-UnixNetVConnection::get_action() const
-{
-  return &action_;
-}
-
 // declarations for local use (within the net module)
 
-void close_UnixNetVConnection(UnixNetVConnection *vc, EThread *t);
 void write_to_net(NetHandler *nh, UnixNetVConnection *vc, EThread *thread);
 void write_to_net_io(NetHandler *nh, UnixNetVConnection *vc, EThread *thread);
